/*
 * 로보타미 스크래치 프로그램 v1.0
 * 개발 시작일 : 2016년 7월 19일
 * 최종 수정일 : 2016년 9월 10일
 * 개발 참여자 : 한성현 (hsh7827@gmail.com)
 * 수정 이력
 *    2016-07-19 = 최초 개발 시작
 */

#include <Servo.h>

#define HEADER1       0x52      // R
#define HEADER2       0x54      // T
#define VER1          0x01      // 메인 버젼 번호
#define VER2          0x00      // 마이너 버젼 번호

/* 명령어 정의문  */
#define STX           0x02      // 패킷의 시작 바이트
#define ETX           0x03      // 패킷의 종료 바이트

#define CMD_OUT       0x41      // 디지털 신호 출력
                                // para1 = 포트, para2 = 신호 값 (0 또는 1)
#define CMD_PWM       0x42      // PWM 신호 출력
                                // para1 = 포트, para2 = PWM 값 (0 ~ 255)
#define CMD_SERVO     0x43      // 서보모터 출력
                                // para1 = 포트, para2 = Servo 값 (0 ~ 180)
#define CMD_DC        0x44      // DC 모터 출력
                                // para1 = 모터 번호, para2 = 모터 속도, para3 = 모터 방향
#define CMD_DC_STOP   0x45      // DC 모터 정지
                                // para1 = 모터 번호
#define CMD_ROBOT     0x46      // 로봇 동작
                                // para1 = 동작 명령, para2 = 동작속도
#define CMD_TONE      0x47      // 톤 (소리)
                                // para1,para2 = 계명, para3, para4 = 지연시간

#define CMD_STATUS    0xF0      // 서버에서 컨트롤러 상태를 요청하는 명령어임.
                                //
#define CMD_REQUEST   0xF1      // 서버에서 현재 입력 핀(디지털, 아날로그) 상태를 요청하는 명령어임.
#define CMD_RESET     0xF2      // 컨트롤러를 리셋함

// ------------------------------------------------------------------
#define CMD_ROBOT_GO         0x01    // 로봇 동작 명령(전진)
#define CMD_ROBOT_BACK       0x02    // 로봇 동작 명령(후진)
#define CMD_ROBOT_LEFT       0x03    // 로봇 동작 명령(좌회전)
#define CMD_ROBOT_RIGHT      0x04    // 로봇 동작 명령(우회전)
#define CMD_ROBOT_LEFT_GO    0x05    // 로봇 동작 명령(전진좌회전)
#define CMD_ROBOT_RIGHT_GO   0x06    // 로봇 동작 명령(전진우회전)
#define CMD_ROBOT_LEFT_TURN  0x07    // 로봇 동작 명령(제자리 좌턴)
#define CMD_ROBOT_RIGHT_TURN 0x08    // 로봇 동작 명령(제자리 우턴)
// ------------------------------------------------------------------

#define CMD_ROBOT_STOP  0x48     // 로봇 정지
#define CMD_TONE_STOP   0x49     // 톤(소리) 정지
#define CMD_TONE_NO     0x4A     // 톤(소리) 지연 없음
                                 // para1,para2 = 계명
#define TONE_PIN         13       // 톤 소리를 낼 포트 지정

// 멜로디 지정 - 옥타브 하나 올라갈 때 마다 2배. B0=31, B1=2*B0, B2=2*B1=4*B0, B3=2*B2=4*B1=8*B0
// Bn = 2^n * B0
#define NOTE_0  0
#define NOTE_C1  33
#define NOTE_CS1 35
#define NOTE_D1  37
#define NOTE_DS1 39
#define NOTE_E1  41
#define NOTE_F1  44
#define NOTE_FS1 46
#define NOTE_G1  49
#define NOTE_GS1 52
#define NOTE_A1  55
#define NOTE_AS1 58
#define NOTE_B1  62

// Define various ADC prescaler
const unsigned char PS_2 = (1 << ADPS0);
const unsigned char PS_4 = (1 << ADPS1);
const unsigned char PS_8 = (1 << ADPS1) | (1 << ADPS0);
const unsigned char PS_16 = (1 << ADPS2);
const unsigned char PS_32 = (1 << ADPS2) | (1 << ADPS0);
const unsigned char PS_64 = (1 << ADPS2) | (1 << ADPS1);
const unsigned char PS_128 = (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0);

/* 포트 형식 정의 상수 */
enum pinType {
  input, output, servomotor, pwm
};

/* 각 포트에 대한 구조체 */
struct pin {
  pinType type;                   // 포트 형식 정의
  int state;                      // 출력 상태
};

pin arduinoPins[14];              // 0-13 포트에 대한 정보
byte value_pinB, value_pinD;      // 디지털 포트 (B, D)에 대한 상태 레지스터를 저장함
unsigned char value_a0, value_a1, value_a2, value_a3, value_a4, value_a5;
                                  // 아날로그 포트(a0 ~ a5) 까지의 값을 저장함
char remainData;

// 설정 함수
//void setup(){
//  configurePins();               // 입/출력 포트 설정
//  resetPins();                   // 입/출력 포트 초기화
//}


// 입/출력 포트 설정
/*
void configurePins() {
  arduinoPins[0].type  = input;           // RX
  arduinoPins[1].type  = output;          // TX
  arduinoPins[2].type  = input;
  arduinoPins[3].type  = input;
  arduinoPins[4].type  = output;          // 모터1 DIR 출력
  arduinoPins[5].type  = pwm;             // 모터1 PWM 출력
  arduinoPins[6].type  = pwm;             // 모터2 PWM 출력
  arduinoPins[7].type  = output;          // 모터2 DIR 출력
  arduinoPins[8].type  = input;
  arduinoPins[9].type  = input;
  arduinoPins[10].type = input;
  arduinoPins[11].type = input;
  arduinoPins[12].type = input;
  arduinoPins[13].type = input;
}
*/

// 입/출력 포트 초기화
/*
void resetPins() {

  delay(10);

  // 각 포트 마다
  for (byte index=0; index <=13; index++) {
    // 포트의 입력 형태에 따라
    switch(arduinoPins[index].type){
      case output:
      case pwm:
        pinMode(index, OUTPUT);
        digitalWrite(index,LOW);
        arduinoPins[index].state=0;
        break;
      case input:
        pinMode(index, INPUT);
        break;
    }
  }
}
*/

// 컨트롤러 리셋
/*
void controller_reset(){
  configurePins();
  resetPins();
}
*/
// 동작 구문
/*
void loop() {
  processingCommand(CMD_ROBOT, CMD_ROBOT_GO, 150,0,0);
  processingCommand(CMD_OUT, 2, 1,0,0);
  processingCommand(CMD_OUT, 3, 1,0,0);
  delay(2000);
  processingCommand(CMD_ROBOT_STOP,0,0,0,0);

  processingCommand(CMD_ROBOT, CMD_ROBOT_LEFT,150,0,0);
  processingCommand(CMD_OUT, 2, 0,0,0);
  processingCommand(CMD_OUT, 3, 0,0,0);
  delay(1000);
  processingCommand(CMD_ROBOT_STOP,0,0,0,0);
}
*/

// 명령어를 처리하는 부분
/*
void processingCommand(byte cmd_byte, byte para1, byte para2, byte para3, byte para4) {
  //명령 코드에 따라 처리
  switch (cmd_byte) {
    // 디지털 신호 출력
    case CMD_OUT:
      // 포트가 출력가 아니면 출력 포트로 만듬
      if (arduinoPins[para1].type != output) make_output(para1);

      // 디지털 신호의 값이 있으면
      if (para2) {
          digitalWrite(para1, HIGH);
          arduinoPins[para1].state = 1;

      // 디지털 신호의 값이 0이면
      } else {
          digitalWrite(para1, LOW);
          arduinoPins[para1].state = 0;
      }
      break;

    // 바퀴형 로봇 동작
    case CMD_ROBOT:
      robot(para1, para2);
      break;
    // 바퀴형 로봇 정지
    case CMD_ROBOT_STOP:
      robot_stop();
      break;
  }
}
*/
// 로봇 동작 함수
void robot(byte robot_cmd, byte robot_speed)
{
  byte speed2;

  speed2 = robot_speed >> 1;

  switch(robot_cmd)
  {
    case CMD_ROBOT_GO:          // 전진
      dc_motor(0, robot_speed, 0);
      dc_motor(1, robot_speed, 1);
      break;
    case CMD_ROBOT_BACK:        // 후진
      dc_motor(0, robot_speed, 1);
      dc_motor(1, robot_speed, 0);
      break;
    case CMD_ROBOT_LEFT:        // 좌회전
      dc_motor(0, 0, 0);
      dc_motor(1, robot_speed, 1);
      break;
    case CMD_ROBOT_RIGHT:       // 우회전
      dc_motor(0, robot_speed, 0);
      dc_motor(1, 0, 1);
      break;
    case CMD_ROBOT_LEFT_GO:     // 전진 좌회전
      dc_motor(0, speed2, 0);
      dc_motor(1, robot_speed, 1);
      break;
    case CMD_ROBOT_RIGHT_GO:    // 전진 우회전
      dc_motor(0, robot_speed, 0);
      dc_motor(1, speed2, 1);
      break;
    case CMD_ROBOT_LEFT_TURN:   // 제자리 좌턴
      dc_motor(0, robot_speed, 1);
      dc_motor(1, robot_speed, 1);
      break;
    case CMD_ROBOT_RIGHT_TURN:  // 제자리 우턴
      dc_motor(0, robot_speed, 0);
      dc_motor(1, robot_speed, 0);
      break;
  }
}

// 로봇 정지 함수
void robot_stop()
{
  dc_motor_stop(2);
}

// DC 모터 정지 함수
void dc_motor_stop(byte motor_no)
{
  switch (motor_no)
  {
    // 1번 모터 정지
    case 0:
      analogWrite(5, 0);
      arduinoPins[5].state = 0;
      break;
    // 2번 모터 정지
    case 1:
      analogWrite(6, 0);
      arduinoPins[6].state = 0;
      break;
    // 1,2 번 모터 모두 정지
    case 2:
      analogWrite(5, 0);
      analogWrite(6, 0);
      arduinoPins[5].state = 0;
      arduinoPins[6].state = 0;
      break;
  }
}

// DC 모터 구동 함수
void dc_motor(byte motor_no, byte motor_pwm, byte motor_dir)
{
  switch (motor_no)
  {
      // 0번 모터는 IO4 - DIR, IO5 - PWM임
      case 0:
        // 포트가 출력가 아니면 출력 포트로 만듬
        if (arduinoPins[4].type != output) make_output(4);
        if (arduinoPins[5].type != pwm) make_pwm(5);

        if (motor_dir) {
          digitalWrite(4, HIGH);
          arduinoPins[4].state = 1;
        } else {
          digitalWrite(4, LOW);
          arduinoPins[4].state = 0;
        }
        analogWrite(5, motor_pwm);
        arduinoPins[5].state = motor_pwm;
        break;
      // 1번 모터는 IO6 - PWM, IO7 - DIR임
      case 1:
        // 포트가 출력가 아니면 출력 포트로 만듬
        if (arduinoPins[7].type != output) make_output(7);
        if (arduinoPins[6].type != pwm) make_pwm(6);

        if (motor_dir) {
          digitalWrite(7, HIGH);
          arduinoPins[7].state = 1;
        } else {
          digitalWrite(7, LOW);
          arduinoPins[7].state = 0;
        }
        analogWrite(6, motor_pwm);
        arduinoPins[6].state = motor_pwm;
        break;

      case 2:
        // 포트가 출력가 아니면 출력 포트로 만듬
        if (arduinoPins[4].type != output) make_output(4);
        if (arduinoPins[5].type != pwm) make_pwm(5);
        if (arduinoPins[7].type != output) make_output(7);
        if (arduinoPins[6].type != pwm) make_pwm(6);

        if (motor_dir) {
          digitalWrite(4, HIGH);
          digitalWrite(7, HIGH);

          arduinoPins[4].state = 1;
          arduinoPins[7].state = 1;
        } else {
          digitalWrite(4, LOW);
          digitalWrite(7, LOW);

          arduinoPins[4].state = 0;
          arduinoPins[7].state = 0;
        }
        analogWrite(5, motor_pwm);
        analogWrite(6, motor_pwm);

        arduinoPins[5].state = motor_pwm;
        arduinoPins[6].state = motor_pwm;

        break;
  }
}

// 포트를 강제적으로 출력으로 만듬
void make_output(byte port_no) {

  arduinoPins[port_no].type = output;
  pinMode(port_no, OUTPUT);

}

// 포트를 강제적으로 PWM으로 만듬
void make_pwm(byte port_no) {

  arduinoPins[port_no].type = pwm;
  pinMode(port_no, OUTPUT);
}

// 포트가 하드웨어 적으로 출력인지 조사
/*
boolean isPortOutput (byte port) {
  if (port > 7) return bitRead(DDRB, port - 8); else return bitRead(DDRD, port);
}

// 입력 포트의 상태를 읽어옴
boolean readPortState (byte port) {
  if (port > 7) return bitRead(PINB, port - 8); else return bitRead(PIND, port);
}

// 현재 디지털 포트의 상태를 업데이트함.
void updateDigitalPort () {
  // 모든 입력 포트에 대해 PIN 레지스터를 읽어들임
  value_pinB = PINB;
  value_pinD = PIND;
}
*/

// 멜로디 연주
void play_melody((int note, int octave, float duration)
{
    tone(13, pow(2,(octave-1))*note, duration);
    delay(duration*1.15);
    noTone(13);
}
